/* all_bz version 10

*  -- generate all blastz commands for pairs of specified sequences.
*
*  The command syntax is:
*	all_bz [-+] string-with-names [blastz-specfile]
*  where "string-with-names" contains names of sequence files (and perhaps
*  non-name characters), such as:
*	all_bz "(((human chimp) baboon) (mouse rat))"
*  A '+' tells all_bz to echo each blastz command to stdout before executing it.
*  A '-' tells all_bz to simply echo the commands (which can be captured in a
*  file, edited, then executed).  The optional last argument contains
*  command-line options for the blastz runs.  For instance, the "specfile":

# The is a sample file specifying command options for blastz
#define MAMMAL human baboon

* : *
	Y=3400
frog : *
	G=11
MAMMAL : zfish
	Q=HoxD50
MAMMAL : MAMMAL
	C=2 B=0
human : zfish
	O=300

*  says e.g. that (1) all blastz runs will specify Y=3400, (2) an alignment
*  involving frog will also specify G=11, (3) an alignment of human and zfish
*  will specify all of: Y=3400 Q=HoxD55 O=300, and (4) a human-vs-baboon
*  alignment will use: Y=3400 C=2 B=0
*
*  CURRENTLY WE ALWAYS SPECIFY: Y=3400 H=2000
*/

#include <stdio.h>
#define RUN_CMD "Blastz %s %s Y=3400 H=2000 %s | lav2maf /dev/stdin %s %s | maf_project /dev/stdin %s | single_cov2 /dev/stdin > %s.%s.maf"

#define VERSION 10

#include "util.h"

struct name_record {
	char *name;
	struct name_record *next;
};

#define NON_NAME " ()"

static char *x[100], *y[100], *z[100];
static int nrules;

int is_comment(char *s) {
	while (*s != '\0' &&isspace(*s))
		++s;
	return (*s == '\0' || (*s == '#' && strncmp(s, "#define ", 8)));
}

char *spec_line(char *buf, int buf_size, FILE *fp) {
	int n = 0;

	while (fgets(buf+n, buf_size-n, fp) != NULL) {
		n = strlen(buf);
		if (n > 1 && buf[n-2] != '\\')
			return buf;
		--n;
		buf[n-1] = ' ';
		buf[n] = '\0';
	}
	return NULL;
}

// parse a blastz specfile; results go into globals x, y, z;
static void get_specs(char *filename) {
	FILE *fp = ckopen(filename, "r");
	char buf[5000], *name[100], *val[100], *s, *t, *u,
	  *v;
	int nmacro = 0, i;

	while (spec_line(buf, 5000, fp) != NULL) {
		if (is_comment(buf))	// blank line
			continue;
		if (strncmp(buf, "#define ", 8) == 0) { // macro definition
			if (nmacro >= 100)
				fatal("too many macros");
			for (s = buf+8; isspace(*s); ++s)
				;
			if ((t = strchr(s, ' ')) == NULL &&
			    (t = strchr(s, '\t')) == NULL)
				fatalf("end of name: %s", buf);
			for (u = t+1; *u != '\0' && isspace(*u); ++u)
				;
			if (*u == '\0')
				fatalf("start of definition: %s", buf);
			if ((v = strchr(u, '\n')) != NULL)
				*v = '\0';
			*t = '\0';
			name[nmacro] = copy_string(s);
			val[nmacro] = copy_string(u);
			++nmacro;
		} else { // must look like " name1 : name2 "
			for (s = buf; *s != '\0'; ++s)
				if (*s == ':' || isspace(*s))
					break;
			if ((u = strchr(s, ':')) == NULL)
				fatalf("needs ':' in %s", buf);
			for (++u; *u != '\0' && isspace(*u); ++u)
				;
			for (v = u; *v != '\0' && !isspace(*v); ++v)
				;
			if (u == v)
				fatalf("confused by %s", buf);
			*s = *v = '\0';
			// name1 (buf)  might be a macro
			for (i = 0; i < nmacro; ++i)
				if (same_string(buf, name[i]))
					break;
			x[nrules] = copy_string(i < nmacro ? val[i] : buf);
			// name2 (u) might be a macro
			for (i = 0; i < nmacro; ++i)
				if (same_string(u, name[i]))
					break;
			y[nrules] = copy_string(i < nmacro ? val[i] : u);
			// next line gives command options
			while (fgets(buf, 500, fp) != NULL && is_comment(buf))
				;
			// it must start with white-space
			if (!isspace(buf[0]))
				fatalf("missing space at start of %s", buf);
			// trim white-space from the front
			for (s = buf+1; isspace(*s); ++s)
				;
			if ((t = strchr(s, '\n')) == NULL)
				fatalf("missing newline in %s", buf);
			// trim white-space from the rear
			while (t > s && isspace(t[-1]))
				--t;
			*t = '\0';
			// record the options in z
			z[nrules] = copy_string(s);
			++nrules;
		}
	}
	fclose(fp);
}

// does name s match pattern t from the blastz-specifile?
static int match(char *s, char *t) {
	return (same_string(t, "*") || strstr(t, s));
}

// put the blastz command options for two species into buf
static void options(char *buf, char *name1, char *name2) {
	int i;

	buf[0] = '\0';
	for (i = 0; i < nrules; ++i)
		if ((match(name1, x[i]) && match(name2, y[i])) ||
		    (match(name1, y[i]) && match(name2, x[i]))) {
			if (buf[0] != '\0')
				strcat(buf, " ");
			strcat(buf, z[i]);
		}
}

int main(int argc, char **argv) {
	struct name_record *m, *n, *names;
	char *s, *t, x, cmd[10000], buf[500];
	int execute = 1, verbose = 0;

	sprintf(cmd, "all_bz.v%d", VERSION);
	argv0 = copy_string(cmd);

	if (argc > 1 && same_string(argv[1], "-")) {
		execute = 0;
		verbose = 1;
		--argc;
		++argv;
	} else if (argc > 1 && same_string(argv[1], "+")) {
		verbose = 1;
		--argc;
		++argv;
	}
	if (argc == 3)
		get_specs(argv[2]);
	else if (argc != 2)
		fatal("-- generate all blastz commands for pairs of specified sequences.\nargs: [-+] species-guid-tree [blastz_specfile]");

	m = names = NULL;
	for (s = argv[1]; *s != '\0'; s = t) {
		while (*s != '\0' && strchr(NON_NAME,*s))
			++s;
		if (*s == '\0')
			break;
		for (t = s+1; *t != '\0' && !strchr(NON_NAME,*t); ++t)
			;
		x = *t;	// save last character
		*t = '\0';
		n = ckalloc(sizeof(*n));
		n->name = copy_string(s);
		*t = x;	// replace last character
		n->next = NULL;
		if (m == NULL)
			names = n;
		else
			m->next = n;
		m = n;
	}
	for (m = names; m != NULL; m = m->next)
	  for (n = m->next; n != NULL; n = n->next) {
	    options(buf, m->name, n->name);
	    if (strstr(buf, "NOALIGN")) {
	      if (verbose)
		printf("do not align %s and %s\n",
		       m->name, n->name);
	      continue;
	    }
	    
	    sprintf(cmd, RUN_CMD, m->name, n->name, buf, m->name, n->name, m->name, m->name, n->name);
	    if ( verbose)
	      printf("%s\n", cmd);
	    if ( execute && system(cmd) != 0)
	      fatal("all_bz quits");
	  }
	return 0;
}
